# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
from typing import Iterator, overload

import numpy as np
import pandas as pd

from lenskit.data import Dataset, ItemListCollection, UserIDKey
from lenskit.data.matrix import MatrixDataset
from lenskit.types import RNGInput
from lenskit.util.random import random_generator

from .split import TTSplit

_log = logging.getLogger(__name__)


def crossfold_records(
    data: Dataset, partitions: int, *, test_only: bool = False, rng: RNGInput = None
) -> Iterator[TTSplit]:
    """
    Partition a dataset by **records** into cross-fold partitions.  This
    partitions the records (ratings, play counts, clicks, etc.) into *k*
    partitions without regard to users or items.

    Since record-based random cross-validation doesn't make much sense with
    repeated interactions, this splitter only supports operating on the
    dataset's interaction matrix.

    Stability:
        Caller

    Args:
        data:
            Ratings or other data you wish to partition.
        partitions:
            The number of partitions to produce.
        test_only:
            If ``True``, returns splits with empty training sets (useful when
            you just want to save the test data).
        rng:
            The random number generator or seed (see :ref:`rng`).

    Returns:
        iterator: an iterator of train-test pairs
    """

    _log.info("partitioning %d ratings into %d partitions", data.count("pairs"), partitions)
    rng = random_generator(rng)

    # get the full data list to split
    df = data.interaction_matrix(format="pandas", field="all", original_ids=True)
    n = len(df)
    rows = np.arange(n)

    # shuffle the indices & split into partitions
    rng.shuffle(rows)
    test_sets = np.array_split(rows, partitions)

    # convert each partition into a split
    for ts in test_sets:
        yield _make_pair(data, df, ts, test_only=test_only)


@overload
def sample_records(
    data: Dataset,
    size: int,
    *,
    disjoint: bool = True,
    test_only: bool = False,
    rng: RNGInput = None,
    repeats: None = None,
) -> TTSplit: ...
@overload
def sample_records(
    data: Dataset,
    size: int,
    *,
    repeats: int,
    disjoint: bool = True,
    test_only: bool = False,
    rng: RNGInput = None,
) -> Iterator[TTSplit]: ...
def sample_records(
    data: Dataset,
    size: int,
    *,
    repeats: int | None = None,
    disjoint: bool = True,
    test_only: bool = False,
    rng: RNGInput = None,
) -> TTSplit | Iterator[TTSplit]:
    """
    Sample train-test a frame of ratings into train-test partitions.  This
    function does not care what kind of data is in `data`, so long as it is a
    Pandas DataFrame (or equivalent).

    We can loop over a sequence of train-test pairs::

        >>> from lenskit.data import load_movielens
        >>> movielens = load_movielens('data/ml-latest-small')
        >>> for split in sample_records(movielens, 1000, repeats=5):
        ...     print(sum(len(il) for il in split.test.lists()))
        1000
        1000
        1000
        1000
        1000

    Sometimes for testing, it is useful to just get a single pair::

        >>> split = sample_records(movielens, 1000)
        >>> sum(len(il) for il in split.test.lists())
        1000

    Stability:
        Caller

    Args:
        data:
            The data set to split.
        size:
            The size of each test sample.
        repeats:
            The number of data splits to produce.  If ``None``, produce a
            _single_ train-test pair instead of an iterator or list.
        disjoint:
            If ``True``, force test samples to be disjoint.
        test_only:
            If ``True``, returns splits with empty training sets (useful when
            you just want to save the test data).
        rng:
            The random number generator or seed (see :ref:`rng`).

    Returns:
        A train-test pair or iterator of such pairs (depending on ``repeats``).
    """

    rng = random_generator(rng)

    # get the full data list to split
    df = data.interaction_matrix(format="pandas", field="all", original_ids=True)
    n = len(df)

    if repeats is None:
        test_pos = rng.choice(np.int32(n), size, replace=False)
        return _make_pair(data, df, test_pos)

    if disjoint and repeats * size >= n:
        _log.warning(
            "wanted %d disjoint splits of %d each, but only have %d rows; cross-folding",
            repeats,
            size,
            n,
        )
        return crossfold_records(data, repeats, rng=rng)

    # get iterators over index arrays for producing the data
    if disjoint:
        _log.info("creating %d disjoint samples of size %d", repeats, size)
        ips = _disjoint_samples(n, size, repeats, rng)

    else:
        _log.info("taking %d samples of size %d", repeats, size)
        ips = _n_samples(n, size, repeats, rng)

    # since this func is both generator and return depending on args,
    # we can't use yield — need to return a generator expression
    return (_make_pair(data, df, test_is, test_only=test_only) for test_is in ips)


def _make_pair(
    data: Dataset,
    df: pd.DataFrame,
    test_is: np.ndarray[int, np.dtype[np.int32]],
    *,
    test_only: bool = False,
) -> TTSplit:
    mask = np.zeros(len(df), np.bool_)
    mask[test_is] = True

    test = ItemListCollection.from_df(df[mask], UserIDKey)
    if test_only:
        train = MatrixDataset(data.users, data.items, df.iloc[:0])
    else:
        train = MatrixDataset(data.users, data.items, df[~mask])

    return TTSplit(train, test)


def _disjoint_samples(
    n: int, size: int, reps: int, rng: np.random.Generator
) -> Iterator[np.ndarray[int, np.dtype[np.int32]]]:
    # shuffle the indices & split into partitions
    xs = np.arange(n, dtype=np.int32)
    rng.shuffle(xs)

    # convert each partition into a split
    for i in range(reps):
        start = i * size
        end = start + size
        yield xs[start:end]


def _n_samples(
    n: int, size: int, reps: int, rng: np.random.Generator
) -> Iterator[np.ndarray[int, np.dtype[np.int32]]]:
    for i in range(reps):
        yield rng.choice(np.int32(n), size, replace=False)



"""
Basic Top-*N* ranking.
"""

import logging

import numpy as np

from lenskit.data import ItemList
from lenskit.pipeline.components import Component

_log = logging.getLogger(__name__)


class TopNRanker(Component):
    """
    Rank scored items by their score and take the top *N*.  The ranking length
    can be passed either at runtime or at component instantiation time, with the
    latter taking precedence.

    Stability:
        Caller

    Args:
        n:
            The desired ranking length.  If negative, then scored items are
            ranked but the ranking is not truncated.
    """

    n: int

    def __init__(self, n: int = -1):
        self.n = n

    def __call__(self, *, items: ItemList, n: int | None = None) -> ItemList:
        """
        Rank the items.

        Args:
            items:
                The items to rank, with scores.  Items with missing scores are
                not included in the final ranking.
            n:
                The number of items to return, or -1 to return all scored items.
                If ``None``, the length configured at construction time is used.

        Returns:
            An ordered list of items, with scores and all other attributes
            preserved.
        """
        if n is None:
            n = self.n

        if n >= 0:
            _log.debug("ranking top %d of %d items", n, len(items))
        else:
            _log.debug("ranking all of %d items", len(items))

        scores = items.scores("numpy")
        if scores is None:
            raise RuntimeError("input item list has no scores")

        # find and filter out invalid scores
        v_mask = ~np.isnan(scores)
        items = items[v_mask]

        # order remaining scores
        order = np.argsort(scores[v_mask])
        if n >= 0 and n < len(order):
            order = order[: -(n + 1) : -1]
        else:
            order = order[::-1]

        # now we need to return in expected order
        result = items[order]
        return ItemList(result, ordered=True)



import logging
from typing import Any

import structlog
from structlog._config import BoundLoggerLazyProxy

_fallback_wrapper = structlog.make_filtering_bound_logger(logging.WARNING)


def get_logger(name: str) -> structlog.stdlib.BoundLogger:
    """
    Get a logger.  This works like :func:`structlog.stdlib.get_logger`, except
    the returned proxy logger is quiet (only WARN and higher messages) if
    structlog has not been configured. LensKit code should use this instead of
    obtaining loggers from Structlog directly.
    """
    return LenskitProxyLogger(None, logger_factory_args=[name])  # type: ignore


class LenskitProxyLogger(BoundLoggerLazyProxy):
    """
    Lazy proxy logger for LensKit.  This is based on Structlog's lazy proxy,
    with using a filtering logger by default when structlog is not configured.
    """

    def bind(self, **new_values: Any):
        if structlog.is_configured():
            self._wrapper_class = None
        else:
            self._wrapper_class = _fallback_wrapper

        return super().bind(**new_values)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Item-based k-NN collaborative filtering.
"""

# pyright: basic
from __future__ import annotations

import warnings

import numpy as np
import torch
from scipy.sparse import csr_array
from typing_extensions import Optional, override

from lenskit import util
from lenskit.data import Dataset, FeedbackType, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.diagnostics import DataWarning
from lenskit.logging import get_logger, trace
from lenskit.logging.progress import item_progress_handle, pbh_update
from lenskit.math.sparse import normalize_sparse_rows, safe_spmv
from lenskit.parallel import ensure_parallel_init
from lenskit.pipeline import Component, Trainable
from lenskit.util.torch import inference_mode

_log = get_logger(__name__)
MAX_BLOCKS = 1024


class ItemKNNScorer(Component, Trainable):
    """
    Item-item nearest-neighbor collaborative filtering feedback. This item-item
    implementation is based on the description of item-based CF by
    :cite:t:`deshpande:iknn` and hard-codes several design decisions found to
    work well in the previous Java-based LensKit code :cite:p:`lenskit-java`. In
    explicit-feedback mode, its output is equivalent to that of the Java
    version.

    .. note::

        This component must be used with queries containing the user's history,
        either directly in the input or by wiring its query input to the output of a
        user history component (e.g., :class:`~lenskit.basic.UserTrainingHistoryLookup`).

    Stability:
        Caller

    Args:
        nnbrs:
            The maximum number of neighbors for scoring each item (``None`` for
            unlimited)
        min_nbrs:
            The minimum number of neighbors for scoring each item
        min_sim:
            Minimum similarity threshold for considering a neighbor.  Must be
            positive; if less than the smallest 32-bit normal (:math:`1.175
            \\times 10^{-38}`), is clamped to that value.
        save_nbrs:
            The number of neighbors to save per item in the trained model
            (``None`` for unlimited)
        feedback:
            The type of input data to use (explicit or implicit).  This affects
            data pre-processing and aggregation.
    """

    nnbrs: int
    min_nbrs: int = 1
    min_sim: float
    save_nbrs: int | None = None
    feedback: FeedbackType
    block_size: int

    items_: Vocabulary
    "Vocabulary of item IDs."
    item_means_: np.ndarray[int, np.dtype[np.float32]] | None
    "Mean rating for each known item."
    item_counts_: np.ndarray[int, np.dtype[np.int32]]
    "Number of saved neighbors for each item."
    sim_matrix_: csr_array
    "Similarity matrix (sparse CSR tensor)."
    users_: Vocabulary
    "Vocabulary of user IDs."

    def __init__(
        self,
        nnbrs: int = 20,
        min_nbrs: int = 1,
        min_sim: float = 1.0e-6,
        save_nbrs: int | None = None,
        feedback: FeedbackType = "explicit",
        block_size: int = 250,
    ):
        self.nnbrs = nnbrs
        self.min_nbrs = min_nbrs
        if self.min_nbrs is not None and self.min_nbrs < 1:
            self.min_nbrs = 1
        self.min_sim = min_sim
        self.save_nbrs = save_nbrs
        self.block_size = block_size

        self.feedback = feedback

        if self.min_sim < 0:
            _log.warning("item-item does not currently support negative similarities")
            warnings.warn("item-item does not currently support negative similarities")
        elif self.min_sim == 0:
            f4i = np.finfo("f4")
            _log.warning(
                "minimum similarity %e is too low, using %e", self.min_sim, f4i.smallest_normal
            )
            self.min_sim = float(f4i.smallest_normal)

    @property
    def is_trained(self) -> bool:
        return hasattr(self, "items_")

    @override
    @inference_mode
    def train(self, data: Dataset):
        """
        Train a model.

        The model-training process depends on ``save_nbrs`` and ``min_sim``, but *not* on other
        algorithm parameters.

        Args:
            ratings:
                (user,item,rating) data for computing item similarities.
        """
        ensure_parallel_init()
        log = _log.bind(n_items=data.item_count, feedback=self.feedback)
        # Training proceeds in 2 steps:
        # 1. Normalize item vectors to be mean-centered and unit-normalized
        # 2. Compute similarities with pairwise dot products
        self._timer = util.Stopwatch()
        log.info("begining IKNN training")

        field = "rating" if self.feedback == "explicit" else None
        init_rmat = data.interaction_matrix("torch", field=field)
        n_items = data.item_count
        log.info(
            "[%s] made sparse matrix",
            self._timer,
            n_ratings=len(init_rmat.values()),
            n_users=data.user_count,
        )

        # we operate on *transposed* rating matrix: items on the rows
        rmat = init_rmat.transpose(0, 1).to_sparse_csr().to(torch.float64)

        if self.feedback == "explicit":
            rmat, means = normalize_sparse_rows(rmat, "center")
            if np.allclose(rmat.values(), 0.0):
                log.warning("normalized ratings are zero, centering is not recommended")
                warnings.warn(
                    "Ratings seem to have the same value, centering is not recommended.",
                    DataWarning,
                )
        else:
            means = None
        log.debug("[%s] centered, memory use %s", self._timer, util.max_memory())

        rmat, _norms = normalize_sparse_rows(rmat, "unit")
        log.debug("[%s] normalized, memory use %s", self._timer, util.max_memory())

        log.info("[%s] computing similarity matrix", self._timer)
        smat = self._compute_similarities(rmat)
        log.debug("[%s] computed, memory use %s", self._timer, util.max_memory())

        log.info(
            "[%s] got neighborhoods for %d of %d items",
            self._timer,
            np.sum(np.diff(smat.crow_indices()) > 0),
            n_items,
        )

        log.info("[%s] computed %d neighbor pairs", self._timer, len(smat.col_indices()))

        self.items_ = data.items
        self.item_means_ = means.numpy() if means is not None else None
        self.item_counts_ = torch.diff(smat.crow_indices()).numpy()
        self.sim_matrix_ = csr_array(
            (smat.values(), smat.col_indices(), smat.crow_indices()), smat.shape
        )
        self.users_ = data.users
        log.debug("[%s] done, memory use %s", self._timer, util.max_memory())

    def _compute_similarities(self, rmat: torch.Tensor) -> torch.Tensor:
        nitems, nusers = rmat.shape

        bs = max(self.block_size, nitems // MAX_BLOCKS)
        _log.debug("computing with effective block size %d", bs)
        with item_progress_handle("items", nitems) as pbh:
            smat = _sim_blocks(rmat.to(torch.float64), self.min_sim, self.save_nbrs, bs, pbh)

        return smat.to(torch.float32)

    @override
    @inference_mode
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        query = RecQuery.create(query)
        log = _log.bind(user_id=query.user_id, n_items=len(items))
        trace(log, "beginning prediction")

        ratings = query.user_items
        if ratings is None or len(ratings) == 0:
            if ratings is None:
                warnings.warn("no user history, did you omit a history component?", DataWarning)
            log.debug("user has no history, returning")
            return ItemList(items, scores=np.nan)

        # set up rating array
        # get rated item positions & limit to in-model items
        ri_nums = ratings.numbers(format="torch", vocabulary=self.items_, missing="negative")
        ri_mask = ri_nums >= 0
        ri_valid_nums = ri_nums[ri_mask]
        n_valid = len(ri_valid_nums)
        trace(log, "%d of %d rated items in model", n_valid, len(ratings))

        if self.feedback == "explicit":
            ri_vals = ratings.field("rating", "numpy")
            if ri_vals is None:
                raise RuntimeError("explicit-feedback scorer must have ratings")
            ri_vals = np.require(ri_vals[ri_mask], np.float32)
        else:
            ri_vals = np.full(n_valid, 1.0, dtype=np.float32)

        # mean-center the rating array
        if self.item_means_ is not None:
            ri_vals -= self.item_means_[ri_valid_nums]

        # convert target item information
        ti_nums = items.numbers(vocabulary=self.items_, missing="negative")
        ti_mask = ti_nums >= 0
        ti_valid_nums = ti_nums[ti_mask]

        # subset the model to rated and target items
        model = self.sim_matrix_
        model = model[ri_valid_nums.numpy(), :]
        assert isinstance(model, csr_array)
        model = model[:, ti_valid_nums]
        assert isinstance(model, csr_array)
        # convert to CSC so we can count neighbors per target item.
        model = model.tocsc()

        # count neighborhood sizes
        sizes = np.diff(model.indptr)
        # which neighborhoods are usable? (at least min neighbors)
        scorable = sizes >= self.min_nbrs

        # fast-path neighborhoods that fit within max neighbors
        fast = sizes <= self.nnbrs
        ti_fast_mask = ti_mask.copy()
        ti_fast_mask[ti_mask] = scorable & fast

        scores = np.full(len(items), np.nan, dtype=np.float32)
        fast_mod = model[:, scorable & fast]
        if self.feedback == "explicit":
            scores[ti_fast_mask] = ri_vals @ fast_mod
            scores[ti_fast_mask] /= fast_mod.sum(axis=0)
        else:
            scores[ti_fast_mask] = fast_mod.sum(axis=0)

        # slow path: neighborhoods that we need to truncate. we will convert to
        # PyTorch, make a dense matrix (this is usually small enough to be
        # usable), and use the Torch topk function.
        slow_mat = model.T[~fast, :]
        assert isinstance(slow_mat, csr_array)
        n_slow, _ = slow_mat.shape
        if n_slow:
            # mask for the slow items.
            ti_slow_mask = ti_mask.copy()
            ti_slow_mask[ti_mask] = ~fast

            slow_mat = torch.from_numpy(slow_mat.toarray())
            slow_trimmed, slow_inds = torch.topk(slow_mat, self.nnbrs)
            assert slow_trimmed.shape == (n_slow, self.nnbrs)
            if self.feedback == "explicit":
                svals = torch.from_numpy(ri_vals)[slow_inds]
                assert svals.shape == slow_trimmed.shape
                scores[ti_slow_mask] = torch.sum(slow_trimmed * svals, axis=1).numpy()
                scores[ti_slow_mask] /= torch.sum(slow_trimmed, axis=1).numpy()
            else:
                scores[ti_slow_mask] = torch.sum(slow_trimmed, axis=1).numpy()

        # re-add the mean ratings in implicit feedback
        if self.item_means_ is not None:
            scores[ti_mask] += self.item_means_[ti_valid_nums]

        log.debug(
            "scored %d items",
            int(np.isfinite(scores).sum()),
        )

        return ItemList(items, scores=scores)

    def __str__(self):
        return "ItemItem(nnbrs={}, msize={})".format(self.nnbrs, self.save_nbrs)


@torch.jit.script
def _sim_row(
    item: int, matrix: torch.Tensor, row: torch.Tensor, min_sim: float, max_nbrs: Optional[int]
) -> tuple[int, torch.Tensor, torch.Tensor]:
    nitems, nusers = matrix.shape
    if len(row.indices()) == 0:
        return 0, torch.zeros((0,), dtype=torch.int32), torch.zeros((0,), dtype=torch.float64)

    # _item_dbg(item, f"comparing item with {row.indices().shape[1]} users")
    # _item_dbg(item, f"row norm {torch.linalg.vector_norm(row.values()).item()}")
    row = row.to_dense()
    sim = safe_spmv(matrix, row.to(torch.float64))
    sim[item] = 0

    mask = sim >= min_sim
    cols = torch.nonzero(mask)[:, 0].to(torch.int32)
    vals = sim[mask]
    # _item_dbg(item, f"found {len(vals)} acceptable similarities")
    assert len(cols) == torch.sum(mask)

    if max_nbrs is not None and max_nbrs > 0 and max_nbrs < vals.shape[0]:
        # _item_dbg(item, "truncating similarities")
        vals, cis = torch.topk(vals, max_nbrs, sorted=False)
        cols = cols[cis]
        order = torch.argsort(cols)
        cols = cols[order]
        vals = vals[order]

    return len(cols), cols, torch.clamp(vals, -1, 1)


@torch.jit.script
def _sim_block(
    matrix: torch.Tensor, start: int, end: int, min_sim: float, max_nbrs: Optional[int], pbh: str
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    "Compute a single block of the similarity matrix"
    bsize = end - start

    counts = torch.zeros(bsize, dtype=torch.int32)
    columns = []
    values = []

    for i in range(start, end):
        c, cs, vs = _sim_row(i, matrix, matrix[i], min_sim, max_nbrs)
        counts[i - start] = c
        columns.append(cs)
        values.append(vs)
        pbh_update(pbh, 1)

    return counts, torch.cat(columns), torch.cat(values).to(torch.float32)


@torch.jit.script
def _sim_blocks(
    matrix: torch.Tensor, min_sim: float, max_nbrs: Optional[int], block_size: int, pbh: str
) -> torch.Tensor:
    "Compute the similarity matrix with blocked matrix-matrix multiplies"
    nitems, nusers = matrix.shape

    jobs: list[torch.jit.Future[tuple[torch.Tensor, torch.Tensor, torch.Tensor]]] = []

    for start in range(0, nitems, block_size):
        end = min(start + block_size, nitems)
        jobs.append(torch.jit.fork(_sim_block, matrix, start, end, min_sim, max_nbrs, pbh))  # type: ignore

    counts = [torch.tensor([0], dtype=torch.int32)]
    columns = []
    values = []

    for job in jobs:
        cts, cis, vs = job.wait()
        counts.append(cts)
        columns.append(cis)
        values.append(vs)

    c_cat = torch.cat(counts)
    crow_indices = torch.cumsum(c_cat, 0, dtype=torch.int32)
    assert len(crow_indices) == nitems + 1
    col_indices = torch.cat(columns)
    c_values = torch.cat(values)
    assert crow_indices[nitems] == len(col_indices)
    assert crow_indices[nitems] == len(c_values)

    return torch.sparse_csr_tensor(
        crow_indices=crow_indices,
        col_indices=col_indices,
        values=c_values,
        size=(nitems, nitems),
    )



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Basic data types used in data representations.
"""

# pyright: strict
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Generic, Literal, NamedTuple, Sequence, TypeAlias, TypedDict, TypeVar, cast

import numpy as np
import pandas as pd

FeedbackType: TypeAlias = Literal["explicit", "implicit"]
"Types of feedback supported."

CoreID: TypeAlias = int | str | bytes
"Core (non-NumPy) identifier types."
NPID: TypeAlias = np.integer[Any] | np.str_ | np.bytes_ | np.object_
"NumPy entity identifier types."
ID: TypeAlias = CoreID | NPID
"Allowable identifier types."
IDArray: TypeAlias = np.ndarray[tuple[int], np.dtype[NPID]]
"NumPy arrays of identifiers."
IDSequence: TypeAlias = Sequence[ID] | IDArray | "pd.Series[CoreID]"
"Sequences of identifiers."

T = TypeVar("T")


@dataclass
class AliasedColumn:
    """
    A data frame column with possible aliases.

    Stability:
        Testing
    """

    name: str
    "The column name."
    compat_aliases: list[str] = field(default_factory=list)
    "A list of aliases for the column."
    warn: bool = False
    "Whether to warn when using an alias."


Column: TypeAlias = str | AliasedColumn


class UIDict(TypedDict, Generic[T]):
    user: T
    item: T


class UITuple(NamedTuple, Generic[T]):
    """
    Tuple of (user, item) data, typically for configuration and similar
    purposes.

    Stability:
        Caller
    """

    user: T
    "User data."
    item: T
    "Item data."

    @classmethod
    def create(cls, x: UITuple[T] | tuple[T, T] | UIDict[T] | T) -> UITuple[T]:
        """
        Create a user-item tuple from a tuple or data.  If a single value
        is provided, it is used for both user and item.
        """
        if isinstance(x, UITuple):
            return cast(UITuple[T], x)
        elif isinstance(x, (tuple, list)):
            u, i = cast(tuple[T, T], x)
            return UITuple(u, i)
        elif isinstance(x, dict):
            return UITuple(x["user"], x["item"])
        else:
            return UITuple(x, x)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit dataset abstraction.
"""

# pyright: basic
from __future__ import annotations

import logging
from typing import (
    Collection,
    Iterable,
    Literal,
    Optional,
    TypeAlias,
    TypeVar,
)

import pandas as pd

from .dataset import Dataset
from .matrix import MatrixDataset
from .types import ID, IDSequence
from .vocab import Vocabulary

DF_FORMAT: TypeAlias = Literal["numpy", "pandas", "torch"]
MAT_FORMAT: TypeAlias = Literal["scipy", "torch", "pandas", "structure"]
MAT_AGG: TypeAlias = Literal["count", "sum", "mean", "first", "last"]
LAYOUT: TypeAlias = Literal["csr", "coo"]
ACTION_FIELDS: TypeAlias = Literal["ratings", "timestamps"] | str

K = TypeVar("K")
_log = logging.getLogger(__name__)


def from_interactions_df(
    df: pd.DataFrame,
    *,
    user_col: str | None = None,
    item_col: str | None = None,
    rating_col: str | None = None,
    timestamp_col: str | None = None,
    users: IDSequence | pd.Index | Iterable[ID] | Vocabulary | None = None,
    items: IDSequence | pd.Index | Iterable[ID] | Vocabulary | None = None,
) -> Dataset:
    """
    Create a dataset from a data frame of ratings or other user-item
    interactions.

    .. todo::
        Repeated interactions are not yet supported.

    Stability:
        Caller
    Args:
        df:
            The user-item interactions (e.g. ratings).  The dataset code takes
            ownership of this data frame and may modify it.
        user_col:
            The name of the user ID column.  By default, looks for columns named
            ``user``, ``user_id``, or ``userId``, with several case variants.
        item_col:
            The name of the item ID column.  By default, looks for columns named
            ``item``, ``item_id``, or ``itemId``, with several case variants.
        rating_col:
            The name of the rating column.
        timestamp_col:
            The name of the timestamp column.
        user_ids:
            A vocabulary of user IDs.  The data frame is subset to this set of IDs.
        item_ids:
            A vocabulary of item IDs.  The data frame is subset to this set of IDs.
    Returns:
        The initiated data set.
    """
    _log.info("creating data set from %d x %d data frame", len(df.columns), len(df))
    df = normalize_interactions_df(
        df,
        user_col=user_col,
        item_col=item_col,
        rating_col=rating_col,
        timestamp_col=timestamp_col,
    )

    if users is not None:
        if not isinstance(users, Vocabulary):
            users = Vocabulary(users, "user")
        df = df[df["user_id"].isin(users.index)]

    if items is not None:
        if not isinstance(items, Vocabulary):
            items = Vocabulary(items, "item")
        df = df[df["item_id"].isin(items.index)]

    df = df.sort_values(["user_id", "item_id"])

    if users is None:
        users = Vocabulary(df["user_id"], "user")

    if items is None:
        items = Vocabulary(df["item_id"], "item")

    return MatrixDataset(users, items, df)


def normalize_interactions_df(
    df: pd.DataFrame,
    *,
    user_col: Optional[str] = None,
    item_col: Optional[str] = None,
    rating_col: Optional[str] = None,
    timestamp_col: Optional[str] = None,
) -> pd.DataFrame:
    """
    Normalize the column names and layout for an interaction data frame.
    """
    _log.debug("normalizing data frame with columns %s", df.columns)
    if user_col is None:
        user_col = _find_column(
            df.columns,
            ["user_id", "user", "USER", "userId", "UserId"],
        )
    if user_col is None:  # pragma nocover
        raise ValueError("no user column found")
    if item_col is None:
        item_col = _find_column(
            df.columns,
            ["item_id", "item", "ITEM", "itemId", "ItemId"],
        )
    if item_col is None:  # pragma nocover
        raise ValueError("no item column found")
    if rating_col is None:
        rating_col = _find_column(
            df.columns,
            ["rating", "RATING"],
        )
    if timestamp_col is None:
        timestamp_col = _find_column(
            df.columns,
            ["timestamp", "TIMESTAMP"],
        )

    _log.debug("id columns: user=%s, item=%s", user_col, item_col)
    _log.debug("rating column: %s", rating_col)
    _log.debug("timestamp column: %s", timestamp_col)

    # rename and reorder columns
    known_columns = ["user_id", "item_id", "rating", "timestamp", "count"]
    renames = {user_col: "user_id", item_col: "item_id"}
    if rating_col:
        renames[rating_col] = "rating"
    if timestamp_col:
        renames[timestamp_col] = "timestamp"
    df = df.rename(columns=renames)
    kc = [c for c in known_columns if c in df.columns]
    oc = [c for c in df.columns if c not in known_columns]
    _log.debug("final columns: %s", kc + oc)
    return df[kc + oc]  # type: ignore


def _find_column(columns: Collection[str], acceptable: Iterable[str]) -> str | None:
    for col in acceptable:
        if col in columns:
            return col

    return None



"""
Basic and baseline pipeline components.
"""

from .bias import BiasModel, BiasScorer
from .candidates import AllTrainingItemsCandidateSelector, UnratedTrainingItemsCandidateSelector
from .composite import FallbackScorer
from .history import UserTrainingHistoryLookup
from .popularity import PopScorer
from .random import RandomSelector, SoftmaxRanker
from .topn import TopNRanker

__all__ = [
    "BiasModel",
    "BiasScorer",
    "PopScorer",
    "TopNRanker",
    "RandomSelector",
    "SoftmaxRanker",
    "UserTrainingHistoryLookup",
    "UnratedTrainingItemsCandidateSelector",
    "AllTrainingItemsCandidateSelector",
    "FallbackScorer",
]



"""
Measure resource consumption.
"""

# pyright: strict
from __future__ import annotations

import os
import time
from dataclasses import dataclass
from pathlib import Path

import torch

from ._proxy import get_logger

_log = get_logger(__name__)


@dataclass
class ResourceMeasurement:
    """
    Single measurement of resources.  Two measurements can be subtracted to
    compute the time resources consumed in an interval (memory resources are
    left unchanged).

    Stability:
        Internal
    """

    wall_time: float
    perf_time: float
    user_time: float
    system_time: float

    max_rss: int | None = None
    """
    Maximum RSS usage (in bytes).
    """
    max_gpu: int | None = None

    @classmethod
    def current(cls):
        """
        Get the current resource measurements.
        """
        wall = time.time()
        perf = time.perf_counter()
        rss = None
        gpu = None
        try:
            import resource

            ru = resource.getrusage(resource.RUSAGE_SELF)
            user = ru.ru_utime
            system = ru.ru_stime
            rss = ru.ru_maxrss * 1024
        except ImportError:
            ts = os.times()
            user = ts.user
            system = ts.system

        if torch.cuda.is_available():
            gpu = torch.cuda.max_memory_allocated()

        return cls(wall, perf, user, system, rss, gpu)

    @property
    def cpu_time(self) -> float:
        "Total CPU time (user + system)."
        return self.user_time + self.system_time

    def __sub__(self, other: ResourceMeasurement):
        return ResourceMeasurement(
            self.wall_time - other.wall_time,
            self.perf_time - other.perf_time,
            self.user_time - other.user_time,
            self.system_time - other.system_time,
            self.max_rss,
        )


def reset_linux_hwm():
    pid = os.getpid()
    reset_file = Path(f"/proc/{pid}/clear_refs")
    if reset_file.exists():
        try:
            reset_file.write_text("5")
        except IOError:
            _log.warn("cannot clear refs", pid=pid)

    if torch.cuda.is_available():
        torch.cuda.reset_peak_memory_stats()



"""
Logging, progress, and resource records.
"""

from ._proxy import get_logger
from .config import LoggingConfig, basic_logging
from .progress import Progress, item_progress, set_progress_impl
from .tasks import Task
from .tracing import trace

__all__ = [
    "LoggingConfig",
    "basic_logging",
    "Progress",
    "item_progress",
    "set_progress_impl",
    "Task",
    "get_logger",
    "trace",
]



"""
LensKit logging processors and converters.
"""

import multiprocessing as mp
from datetime import datetime
from typing import Any

import structlog
from structlog.typing import EventDict


def remove_internal(logger: Any, method: str, event_dict: EventDict) -> EventDict:
    """
    Filter out “internal” attrs (beginning with ``_``) for console logging.
    """

    to_del = [k for k in event_dict.keys() if k.startswith("_")]
    for k in to_del:
        del event_dict[k]

    return event_dict


def format_timestamp(logger: Any, method: str, event_dict: EventDict) -> EventDict:
    """
    Reformat UNIX timestamps.
    """

    if "timestamp" in event_dict:
        stamp = datetime.fromtimestamp(event_dict["timestamp"])
        event_dict = dict(event_dict)
        event_dict["timestamp"] = stamp.isoformat(timespec="seconds")
        return event_dict
    else:
        return event_dict


def add_process_info(logger: Any, method: str, event_dict: EventDict) -> EventDict:
    """
    Add process info if it does not exist.
    """

    proc = mp.current_process()
    if "pid" not in event_dict:
        event_dict["pid"] = proc.pid
    if "pname" not in event_dict:
        event_dict["pname"] = proc.name

    return event_dict


def log_warning(message, category, filename, lineno, file=None, line=None):
    log = structlog.stdlib.get_logger("lenskit")
    log = log.bind(category=category.__name__, file=filename, lineno=lineno)
    log.warning(str(message))



"""
Recommender systems toolkit.
"""

__path__ = __import__("pkgutil").extend_path(__path__, __name__)

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "UNSPECIFIED"


from .operations import predict, recommend, score
from .pipeline import Pipeline, RecPipelineBuilder, topn_pipeline

__all__ = ["predict", "recommend", "score", "Pipeline", "RecPipelineBuilder", "topn_pipeline"]



from __future__ import annotations

import warnings
from collections import namedtuple
from collections.abc import Sequence
from os import PathLike
from pathlib import Path
from typing import (
    Any,
    Generator,
    Generic,
    Iterator,
    Literal,
    Mapping,
    NamedTuple,
    TypeAlias,
    TypeVar,
    overload,
)

import pandas as pd
import pyarrow as pa
from more_itertools import chunked
from pyarrow.parquet import ParquetDataset, ParquetWriter

from lenskit.diagnostics import DataWarning

from .items import ItemList
from .schemas import column_name, normalize_columns
from .types import ID, Column

K = TypeVar("K", bound=tuple)
KeySchema: TypeAlias = type[K] | tuple[str, ...]
GenericKey: TypeAlias = tuple[ID, ...]


class UserIDKey(NamedTuple):
    """
    Key type for user IDs.  This is used for :class:`item list collections
    <ItemListCollection>` that are keyed by user ID, a common setup for
    recommendation runs and
    """

    user_id: ID


KEY_CACHE: dict[tuple[str, ...], type[tuple]] = {("user_id",): UserIDKey}


class ItemListCollection(Generic[K]):
    """
    A collection of item lists.

    An item list collection consists of a sequence of item lists with associated
    *keys* following a fixed schema.  Item list collections support iteration
    (in order) and lookup by key. They are used to represent a variety of
    things, including test data and the results of a batch run.

    The key schema can be specified either by a list of field names, or by
    providing a named tuple class (created by either :func:`namedtuple` or
    :class:`NamedTuple`) defining the key schema.  Schemas should **not** be
    nested: field values must be scalars, not tuples or lists.  Keys should also
    be hashable.

    This class exists, instead of using raw dictionaries or lists, to
    consistently handle some of the nuances of multi-valued keys, and different
    collections having different key fields; for example, if a run produces item
    lists with both user IDs and sequence numbers, but your test data is only
    indexed by user ID, the *projected lookup* capabilities make it easy to find
    the test data to go with an item list in the run.

    Item list collections support lookup by index, like a list, returning a
    tuple of the key and list.  If they are constructed with ``index=True``,
    they also support lookup by _key_, supplied as either a tuple or an instance
    of the key type; in this case, the key is not returned.  If more than one
    item with the same key is inserted into the collection, then the _last_ one
    is returned (just like a dictionary).

    Args:
        key:
            The type (a NamedTuple class) or list of field names specifying the
            key schema.
        index:
            Whether or not to index lists by key to facilitate fast lookups.
    """

    _key_class: type[K]
    _lists: list[tuple[K, ItemList]]
    _index: dict[K, int] | None = None
    _list_schema: dict[str, pa.DataType]

    def __init__(self, key: type[K] | Sequence[str], *, index: bool = True):
        """
        Create a new item list collection.
        """
        if isinstance(key, type):
            self._key_class = key
        else:
            self._key_class = _create_key_type(*key)  # type: ignore

        self._lists = []
        if index:
            self._index = {}
        self._list_schema = {}

    @overload
    @classmethod
    def from_dict(
        cls, data: Mapping[GenericKey | ID, ItemList], key: type[K]
    ) -> ItemListCollection[K]: ...
    @overload
    @classmethod
    def from_dict(
        cls, data: Mapping[GenericKey | ID, ItemList], key: Sequence[str] | str | None = None
    ) -> ItemListCollection[GenericKey]: ...
    @classmethod
    def from_dict(
        cls,
        data: Mapping[GenericKey | ID, ItemList],
        key: type[K] | Sequence[str] | str | None = None,
    ) -> ItemListCollection[Any]:
        """
        Create an item list collection from a dictionary.
        """
        if key is not None:
            if isinstance(key, str):
                key = (key,)
            ilc = ItemListCollection(key)
        else:
            k = next(iter(data.keys()))
            if isinstance(k, tuple) and getattr(k, "_fields"):
                ilc = ItemListCollection(type(k))
            else:
                warnings.warn(
                    "no key specified but data does not use named tuples, using default field 'id'",
                    DataWarning,
                )
                ilc = ItemListCollection(["id"])

        for k, il in data.items():
            if isinstance(k, tuple):
                ilc.add(il, *k)
            else:
                ilc.add(il, k)

        return ilc

    @classmethod
    def from_df(cls, df: pd.DataFrame, key: type[K] | Sequence[Column] | Column, *others: Column):
        """
        Create an item list collection from a data frame.

        .. note::

            Keys with empty item lists will be silently excluded from the output
            data.

        Args:
            df:
                The data frame to convert.
            key:
                The key type or field(s).  Can be specified as a single column
                name (or :class:`~lenskit.data.types.AliasedColumn`).
            others:
                Other columns to consider; primarily used to pass additional
                aliased columns to normalize other clumnes like the item ID.
        """
        if isinstance(key, type):
            fields = _key_fields(key)
            columns = fields + others
        else:
            if isinstance(key, Column):
                key = [key]
            columns = tuple(key) + others
            fields = [column_name(c) for c in key]
            key = _create_key_type(*fields)  # type: ignore

        df = normalize_columns(df, *columns)
        ilc = cls(key)  # type: ignore
        for k, gdf in df.groupby(list(fields)):
            ilc.add(ItemList.from_df(gdf), *k)

        return ilc

    def to_df(self) -> pd.DataFrame:
        """
        Convert this item list collection to a data frame.

        .. warning::

            If this item list collection has any keys with empty lists, those
            lists will be excluded from the output.
        """
        fields = self._key_class._fields  # type: ignore
        if any(len(il) == 0 for il in self.lists()):
            warnings.warn(
                "item list collection has empty lists, they will be dropped",
                DataWarning,
                stacklevel=2,
            )
        return (
            pd.concat({k: il.to_df(numbers=False) for (k, il) in self._lists}, names=fields)
            .reset_index(fields)
            .reset_index(drop=True)
        )

    def to_arrow(self, *, batch_size: int = 5000) -> pa.Table:
        """
        Convert this item list collection to an Arrow table.

        The resulting table has one row per item list, with the item list
        contents an ``items`` column of a structured list type.  This preserves
        empty item lists for higher-fidelity data storage.

        Args:
            batch_size:
                The Arrow record batch size.
        """
        return pa.Table.from_batches(self._iter_record_batches())

    def save_parquet(
        self,
        path: PathLike[str],
        *,
        layout: Literal["native", "flat"] = "native",
        batch_size: int = 5000,
        compression: Literal["zstd", "gzip", "snappy", "lz4"] | None = "zstd",
    ) -> None:
        """
        Save this item list collection to a Parquet file.  This supports two
        types of Parquet files: “native” collections store one row per list,
        with the item list contents in a repeated structure column named
        ``items``; this layout fully preserves the item list collection,
        including empty item lists.  The “flat” layout is easier to work with in
        software such as Pandas, but cannot store empty item lists.

        Args:
            layout:
                The table layout to use.
            batch_size:
                The Arrow record batch size.
            compression:
                The compression scheme to use.
        """
        if layout == "flat":
            self.to_df().to_parquet(path, compression=compression)
            return

        writer = None
        try:
            for batch in self._iter_record_batches(batch_size):
                if writer is None:
                    writer = ParquetWriter(Path(path), batch.schema, compression=compression)
                writer.write_batch(batch)
        finally:
            if writer is not None:
                writer.close()

    @classmethod
    def load_parquet(cls, path: PathLike[str] | list[PathLike[str]]) -> ItemListCollection:
        """
        Load this item list from a Parquet file using the native layout.

        .. note::

            To load item list collections in the flat layout, use Pandas and
            :meth:`from_df`.

        Args:
            path:
                Path to the Parquet file to load.
        """
        if isinstance(path, list):
            path = [Path(p) for p in path]
        else:
            path = Path(path)
        dataset = ParquetDataset(path)  # type: ignore
        table = dataset.read()
        keys = table.drop("items")
        lists = table.column("items")
        ilc = ItemListCollection(keys.schema.names)
        for i, key in enumerate(keys.to_pylist()):
            il_data = lists[i].values
            ilc.add(ItemList.from_arrow(il_data), **key)

        return ilc

    def _iter_record_batches(
        self, batch_size: int = 5000, columns: dict[str, pa.DataType] | None = None
    ) -> Generator[pa.RecordBatch, None, None]:
        if columns is None:
            columns = self._list_schema

        for batch in chunked(self._lists, batch_size):
            keys = pa.Table.from_pylist([_key_dict(k) for (k, _il) in batch])
            schema = pa.list_(pa.struct(columns))  # type: ignore
            tbl = keys.add_column(
                keys.num_columns,
                "items",
                pa.array(
                    [il.to_arrow(type="array", columns=columns) for (_k, il) in batch],
                    schema,
                ),
            )
            yield from tbl.to_batches()

    @property
    def key_fields(self) -> tuple[str]:
        "The names of the key fields."
        return _key_fields(self._key_class)

    @property
    def key_type(self) -> type[K]:
        """
        The type of collection keys.
        """
        return self._key_class

    def add(self, list: ItemList, *fields: ID, **kwfields: ID):
        """
        Add a single item list to this list.

        Args:
            list:
                The item list to add.
            fields, kwfields:
                The key fields for this list.
        """
        key = self._key_class(*fields, **kwfields)  # type: ignore
        self._add(key, list)

    def add_from(self, other: ItemListCollection, **fields: ID):
        """
        Add all collection from another collection to this collection.  If field
        values are supplied, they are used to supplement or overwrite the keys
        in ``other``; a common use case is to add results from multiple
        recommendation runs and save them a single field.

        Args:
            other:
                The item list collection to incorporate into this one.
            fields:
                Additional key fields (must be specified by name).
        """
        for key, list in other:
            if fields:
                cf = key._asdict() | fields
                key = self._key_class(**cf)
            self._add(key, list)

    def _add(self, key: K, list: ItemList):
        self._lists.append((key, list))
        if self._index is not None:
            self._index[key] = len(self._lists) - 1
        for fn, ft in list.arrow_types().items():
            pft = self._list_schema.get(fn, None)
            if pft is None:
                self._list_schema[fn] = ft
            elif not ft.equals(pft):
                raise TypeError(f"incompatible item lists: field {fn} type {ft} != {pft}")

    @overload
    def lookup(self, key: tuple) -> ItemList | None: ...
    @overload
    def lookup(self, *key: ID, **kwkey: ID) -> ItemList | None: ...
    def lookup(self, *args, **kwargs) -> ItemList | None:
        """
        Look up a list by key.  If multiple lists have the same key, this
        returns the **last** (like a dictionary).

        This method can be called with the key tuple as a single argument (and
        this can be either the actual named tuple, or an ordinary tuple of the
        same length), or with the individual key fields as positional or named
        arguments.

        Args:
            key:
                The key tuple or key tuple fields.
        """
        if self._index is None:
            raise TypeError("cannot lookup on non-indexed collection")
        if len(args) != 1 or not isinstance(args[0], tuple):
            key = self._key_class(*args, **kwargs)
        else:
            key = args[0]

        try:
            return self._lists[self._index[key]][1]  # type: ignore
        except KeyError:
            return None

    def lookup_projected(self, key: tuple) -> ItemList | None:
        """
        Look up an item list using a *projected* key.  A projected key is a key
        that may have additional fields beyond those defined by this collection,
        that are ignored for the purposes of lookup.

        Args:
            key:
                The key.  Must be a named tuple (e.g. a key obtained from
                another item list collection).

        Returns:
            The item list with the specified key, projected to this collection's
            key fields, or ``None`` if no such list exists.
        """
        kp = project_key(key, self._key_class)
        return self.lookup(kp)

    def lists(self) -> Iterator[ItemList]:
        "Iterate over item lists without keys."
        return (il for (_k, il) in self._lists)

    def keys(self) -> Iterator[K]:
        "Iterate over keys."
        return (k for (k, _il) in self._lists)

    def __len__(self):
        return len(self._lists)

    def __iter__(self) -> Iterator[tuple[K, ItemList]]:
        return iter(self._lists)

    def __getitem__(self, key: int) -> tuple[K, ItemList]:
        return self._lists[key]


def _key_fields(kt: type[tuple]) -> tuple[str]:
    "extract the fields from a key type"
    return kt._fields  # type: ignore


def _key_dict(kt: tuple[ID, ...]) -> Mapping[str, Any]:
    return kt._asdict()  # type: ignore


@overload
def _create_key(kt: type[K], *values: ID) -> K: ...
@overload
def _create_key(kt: Sequence[str], *values: ID) -> GenericKey: ...
def _create_key(kt: type[K] | Sequence[str], *values: ID) -> tuple[Any, ...]:
    if isinstance(kt, type):
        return kt(*values)  # type: ignore
    else:
        kt = _create_key_type(*kt)  # type: ignore
        return kt(*values)  # type: ignore


def _create_key_type(*fields: str) -> type[GenericKey]:
    """
    Create a new key
    """
    assert isinstance(fields, tuple)
    kt = KEY_CACHE.get(fields, None)
    if kt is None:
        ktn = f"LKILCKeyType{len(KEY_CACHE)+1}"
        kt = namedtuple(ktn, fields)
        # support pickling
        kt.__reduce__ = _reduce_generic_key  # type: ignore
        KEY_CACHE[fields] = kt
    return kt


def _reduce_generic_key(key):
    args = (key._fields,) + key
    return _create_key, args


def project_key(key: tuple, target: type[K]) -> K:
    """
    Project a key onto a subset of its fields.  This is to enable keys to be
    looked up in other collections that are keyed on a subset of their fields,
    such as using a key consisting of a user ID and a sequence number to look up
    test data in a collection keyed only by user ID.
    """

    if isinstance(key, target):
        return key

    try:
        return target._make(getattr(key, f) for f in target._fields)  # type: ignore
    except AttributeError as e:
        raise TypeError(f"source key is missing field {e.name}")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Efficient solver routines.
"""

import torch


def solve_cholesky(A: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """
    Solve the system :math:`A\\mathbf{x}=\\mathbf{y}` for :math:`\\mathbf{x}`
    with Cholesky decomposition.

    This wraps :func:`torch.linalg.cholesky_ex` and :func:`torch.cholesky_solve`
    in an easier-to-use interface with error checking.

    Args:
        A:
            the left-hand matrix :math:`A`
        y:
            the right-hand vector :math:`\\mathbf{y}`

    Returns:
        the solution :math:`\\mathbf{x}`
    """
    if len(y.shape) > 1:  # pragma: no cover
        raise TypeError(f"y must be 1D (found shape {y.shape})")
    (n,) = y.shape
    if A.shape != (n, n):  # pragma: no cover
        raise TypeError("A must be n⨉n")

    L, info = torch.linalg.cholesky_ex(A)
    if info:
        raise RuntimeError("cholesky solve failed")
    y = y.reshape(1, n, 1)
    return torch.cholesky_solve(y, L).reshape(n)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT
from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, Literal, TypeAlias, TypeVar

import pandas as pd

from lenskit.data import Dataset, ItemListCollection
from lenskit.data.matrix import MatrixDataset

SplitTable: TypeAlias = Literal["matrix"]
TK = TypeVar("TK", bound=tuple)


@dataclass
class TTSplit(Generic[TK]):
    """
    A train-test set from splitting or other sources.

    Stability:
        Caller
    """

    train: Dataset
    """
    The training data.
    """

    test: ItemListCollection[TK]
    """
    The test data.
    """

    @property
    def test_size(self) -> int:
        """
        Get the number of test pairs.
        """
        return sum(len(il) for il in self.test.lists())

    @property
    def test_df(self) -> pd.DataFrame:
        """
        Get the test data as a data frame.
        """
        return self.test.to_df()

    @property
    def train_df(self) -> pd.DataFrame:
        """
        Get the training data as a data frame.
        """
        return self.train.interaction_matrix("pandas", field="all")

    @classmethod
    def from_src_and_test(cls, src: Dataset, test: ItemListCollection[TK]) -> TTSplit[TK]:
        """
        Create a split by subtracting test data from a source dataset.
        """
        cols = list(test.key_fields) + ["item_id"]

        test_df = test.to_df().set_index(cols)
        mask = pd.Series(False, index=test_df.index)

        df = src.interaction_matrix("pandas", field="all", original_ids=True).set_index(cols)

        mask = mask.reindex(df.index, fill_value=True)

        train_df = df[mask]
        train = MatrixDataset(src.users, src.items, train_df.reset_index())

        return cls(train, test)



"""
LensKit test harnesses and utilities.

This package contains utility code for testing LensKit and its components,
including in derived packages.  It relies on PyTest and Hypothesis.
"""

import os
from contextlib import contextmanager

from ._arrays import coo_arrays, scored_lists, sparse_arrays, sparse_tensors
from ._components import BasicComponentTests, ScorerTests
from ._markers import jit_enabled, wantjit
from ._movielens import (
    demo_recs,
    ml_100k,
    ml_100k_zip,
    ml_ds,
    ml_ds_unchecked,
    ml_ratings,
    ml_test_dir,
)

__all__ = [
    "coo_arrays",
    "scored_lists",
    "sparse_arrays",
    "sparse_tensors",
    "ml_100k",
    "ml_100k_zip",
    "ml_ds",
    "ml_ds_unchecked",
    "ml_ratings",
    "ml_test_dir",
    "demo_recs",
    "wantjit",
    "jit_enabled",
    "set_env_var",
    "BasicComponentTests",
    "ScorerTests",
]


@contextmanager
def set_env_var(var, val):
    "Set an environment variable & restore it."
    old_val = os.environ.get(var, None)
    try:
        if val is None:
            if old_val is not None:
                del os.environ[var]
        else:
            os.environ[var] = val
        yield
    finally:
        if old_val is not None:
            os.environ[var] = old_val
        elif val is not None:
            del os.environ[var]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Serialization utilities for parallel processing.
"""

import io
import logging
import pickle
from multiprocessing.managers import SharedMemoryManager
from multiprocessing.shared_memory import SharedMemory
from typing import Any, NamedTuple

import torch
from torch.multiprocessing.reductions import reduce_storage, reduce_tensor

_log = logging.getLogger(__name__)


class SHMData(NamedTuple):
    """
    Serialized data (with shared memory handles).
    """

    pickle: bytes
    buffers: list[tuple[SharedMemory | None, int]]


class SHMPickler(pickle.Pickler):
    manager: SharedMemoryManager | None
    buffers: list[tuple[SharedMemory | None, int]]

    def __init__(
        self,
        file,
        protocol: int | None = pickle.HIGHEST_PROTOCOL,
        manager: SharedMemoryManager | None = None,
        *,
        fix_imports: bool = False,
    ) -> None:
        super().__init__(file, protocol, fix_imports=fix_imports, buffer_callback=self._buffer_cb)
        self.manager = manager
        self.buffers = []

    def _buffer_cb(self, buffer: pickle.PickleBuffer):
        mem = buffer.raw()
        if mem.nbytes == 0:
            shm = None
        elif self.manager:
            shm = self.manager.SharedMemory(mem.nbytes)
        else:
            shm = SharedMemory(create=True, size=mem.nbytes)

        # copy the data
        if shm is not None:
            shm.buf[: mem.nbytes] = mem
        self.buffers.append((shm, mem.nbytes))

    def reducer_override(self, obj: Any) -> Any:
        if isinstance(obj, torch.Tensor):
            if obj.is_sparse_csr:
                return torch.sparse_csr_tensor, (
                    obj.crow_indices(),
                    obj.col_indices(),
                    obj.values(),
                    obj.shape,
                )
            elif obj.is_sparse:
                return torch.sparse_coo_tensor, (
                    obj.indices(),
                    obj.values(),
                    obj.shape,
                )
            elif obj.layout == torch.sparse_csc:
                return torch.sparse_csc_tensor, (
                    obj.ccol_indices(),
                    obj.row_indices(),
                    obj.values(),
                    obj.shape,
                )
            else:
                return reduce_tensor(obj)

        if isinstance(obj, torch.UntypedStorage):
            return reduce_storage(obj)

        return NotImplemented


def shm_serialize(obj: Any, manager: SharedMemoryManager | None = None) -> SHMData:
    """
    Serialize an object for processing in a subclass with shared memory when
    feasible (including CUDA).
    """
    out = io.BytesIO()
    pkl = SHMPickler(out, pickle.HIGHEST_PROTOCOL, manager)
    pkl.dump(obj)

    data = out.getvalue()
    _log.debug("serialized model into %s pickle bytes", len(data))
    return SHMData(bytes(data), pkl.buffers)


def shm_deserialize(data: SHMData) -> Any:
    """
    Deserialize SHM-pickled data.
    """
    buffers = [(shm.buf[:n] if shm is not None else b"") for shm, n in data.buffers]
    return pickle.loads(data.pickle, buffers=buffers)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict

import warnings
from inspect import Signature, signature

from typing_extensions import Generic, TypeVar

from .components import PipelineFunction
from .types import TypecheckWarning

# Nodes are (conceptually) immutable data containers, so Node[U] can be assigned
# to Node[T] if U ≼ T.
ND = TypeVar("ND", covariant=True)


class Node(Generic[ND]):
    """
    Representation of a single node in a :class:`Pipeline`.

    Stability:
        Caller
    """

    __match_args__ = ("name",)

    name: str
    "The name of this node."
    types: set[type] | None
    "The set of valid data types of this node, or None for no typechecking."

    def __init__(self, name: str, *, types: set[type] | None = None):
        self.name = name
        self.types = types

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} {self.name}>"


class InputNode(Node[ND], Generic[ND]):
    """
    An input node.

    Stability:
        Internal
    """


class LiteralNode(Node[ND], Generic[ND]):
    """
    A node storing a literal value.

    Stability:
        Internal
    """

    __match_args__ = ("name", "value")
    value: ND
    "The value associated with this node"

    def __init__(self, name: str, value: ND, *, types: set[type] | None = None):
        super().__init__(name, types=types)
        self.value = value


class ComponentNode(Node[ND], Generic[ND]):
    """
    A node storing a component.

    Stability:
        Internal
    """

    __match_args__ = ("name", "component", "inputs", "connections")

    component: PipelineFunction[ND]
    "The component associated with this node"

    inputs: dict[str, type | None]
    "The component's inputs."

    connections: dict[str, str]
    "The component's input connections."

    def __init__(self, name: str, component: PipelineFunction[ND]):
        super().__init__(name)
        self.component = component
        self.connections = {}

        sig = signature(component, eval_str=True)
        if sig.return_annotation == Signature.empty:
            warnings.warn(
                f"component {component} has no return type annotation", TypecheckWarning, 2
            )
        else:
            self.types = set([sig.return_annotation])

        self.inputs = {}
        for param in sig.parameters.values():
            if param.annotation == Signature.empty:
                warnings.warn(
                    f"parameter {param.name} of component {component} has no type annotation",
                    TypecheckWarning,
                    2,
                )
                self.inputs[param.name] = None
            else:
                self.inputs[param.name] = param.annotation



"""
Recommendation queries.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TypeAlias

import numpy as np

from lenskit.data.items import ItemList

from .types import ID


@dataclass
class RecQuery:
    """
    Representation of a the data available for a recommendation query.  This is
    generally the available inputs for a recommendation request *except* the set
    of candidate items.

    .. todo::
        Document and test methods for extending the recommendation query with arbitrary
        data to be used by client-provided pipeline components.

    .. todo::
        When LensKit supports context-aware recommendation, this should be extended
        to include context cues.

    Stability:
        Caller
    """

    user_id: ID | None = None
    """
    The user's identifier.
    """
    user_items: ItemList | None = None
    """
    The items from the user's interaction history, with ratings if available.  This
    list is *deduplicated*, like :meth:`~lenskit.data.Dataset.interaction_matrix`,
    rather than a full interaction log.  As with :meth:`~lenskit.data.Dataset.user_row`,
    the rating is expected to be on a ``rating`` field, not ``score``.
    """

    @classmethod
    def create(cls, data: QueryInput) -> RecQuery:
        """
        Create a recommenadtion query from an input, filling in available
        components based on the data type.

        Args:
            data:
                Input data to turn into a recommendation query.  If the input is
                already a query, it is returned *as-is* (not copied).

        Returns:
            The recommendation query.
        """
        if data is None:
            return RecQuery()
        elif isinstance(data, RecQuery):
            return data
        elif isinstance(data, ItemList):
            return cls(user_items=data)
        elif isinstance(data, np.integer):
            return cls(user_id=data.item())
        elif isinstance(data, int | str | bytes):
            return cls(user_id=data)
        else:  # pragma: nocover
            raise TypeError(f"invalid query input (type {type(data)})")


QueryInput: TypeAlias = RecQuery | ID | ItemList | None
"""
Types that can be converted to a query by :meth:`RecQuery.create`.
"""
